import datetime as dt
from pathlib import Path
from typing import cast

from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.entities import Entity
from ynab_unlinked.exceptions import ParsingError
from ynab_unlinked.models import Transaction
from ynab_unlinked.parsers import pdf


class BBVA(Entity):
    def parse(
        self, input_file: Path, context: YnabUnlinkedContext
    ) -> list[Transaction]:
        transactions = []

        for row in pdf(
            input_file, allow_empty_columns=False, expected_number_of_columns=3
        ):
            parsed_row = self.__extract_fields_from_row(cast(list[str], row))
            if parsed_row is None:
                raise ParsingError(
                    input_file,
                    "Malformed Transaction Table: Could not extract the date, payee and a mount for one row",
                )

            date, payee, amount = parsed_row

            # It can be that the row does not represent a transaction. Skip it
            try:
                parsed_date = dt.datetime.strptime(date, "%d/%m/%Y").date()
            except ValueError:
                continue

            transactions.append(
                Transaction(
                    date=parsed_date,
                    payee=payee,
                    amount=float(amount.replace("€", "").replace(",", ".")),
                )
            )

        return transactions

    def __extract_fields_from_row(self, row: list[str]) -> tuple[str, ...] | None:
        # PDFs should have three columns
        # - Date with 2 lines for the date the transaction took place and when it was approved
        # - Concept, used for payee. Sometimes 2 lines including spending category
        # - Amount
        if len(row) != 3:
            return None

        date = row[0].splitlines()[0]
        payee = row[1].splitlines()[0]
        amount = row[2]

        return date, payee, amount

    def name(self) -> str:
        return "BBVA Credit Card"



import importlib
import pkgutil
from typing import Annotated

import typer

from ynab_unlinked import entities
from ynab_unlinked.context_object import YnabUnlinkedContext

load = typer.Typer(
    help="Load transactions from a bank statement into your YNAB account.",
)


@load.callback()
def load_callback(
    context: typer.Context,
    show: Annotated[
        bool,
        typer.Option(
            "-s",
            "--show",
            help="Just show the transactions available in the input file.",
        ),
    ] = False,
    reconcile: Annotated[
        bool, typer.Option("-r", "--reconcile", help="Reconcile cleared transactions")
    ] = False,
):
    obj: YnabUnlinkedContext = context.obj

    obj.show = show
    obj.reconcile = reconcile


# Dynamically load all entities commands when present
for _finder, name, ispkg in pkgutil.iter_modules(entities.__path__):
    if not ispkg:
        continue

    module = importlib.import_module(f"{entities.__name__}.{name}")
    if not hasattr(module, "command"):
        continue

    command = module.command

    if callable(command):
        load.command(name=name, no_args_is_help=True)(command)



import datetime as dt
from typing import Annotated, Literal

from rich import print
from rich.prompt import Confirm, Prompt
from rich.status import Status
from typer import Context, Option
from ynab.models.transaction_cleared_status import TransactionClearedStatus

from ynab_unlinked import app, display
from ynab_unlinked.config import TRANSACTION_GRACE_PERIOD_DAYS
from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.ynab_api import Client


def indexes_to_reconcile(max: int) -> list[int] | Literal["all"]:
    selection = Prompt.ask(
        (
            "Select which accounts you want to reconcile from the list above. You can suply several "
            "of them with a comma separated list of numbers (e.g. 1,2,4) or answer 'all' to reconcile all acounts."
        ),
        default="all",
    )

    if selection == "all":
        return "all"

    while True:
        try:
            if selection == ".q":
                return []

            indexes = [int(i.strip()) for i in selection.split(",")]

            if any(i > max for i in indexes):
                raise RuntimeError()

            return indexes
        except Exception:
            selection = Prompt.ask(
                (
                    "Select either a set of numbers (e.g. 1,2,4) or 'all' to reoncile all accounts.\n"
                    "If you want to quit, answer '.q'"
                ),
                default="all",
            )


@app.command()
def reconcile(
    context: Context,
    all: Annotated[
        bool,
        Option(
            "--all",
            "-a",
            is_flag=True,
            help=(
                "Get all transactions. By default, only transactions created after the last "
                "time this command was run will be considered."
            ),
        ),
    ] = False,
    uncleared: Annotated[
        bool,
        Option(
            "--uncleared",
            "-u",
            is_flag=True,
            help="Reconcile even uncleared transactions",
        ),
    ] = False,
):
    """Help reconciling your accounts in one go"""

    ctx: YnabUnlinkedContext = context.obj

    last_reconciliation_date = None if all else ctx.config.last_reconciliation_date

    client = Client(ctx.config)

    cleared_allowed = {TransactionClearedStatus.CLEARED}
    if uncleared:
        cleared_allowed.add(TransactionClearedStatus.UNCLEARED)

    with Status("Getting transactions from YNAB"):
        transactions_to_reconcile = [
            transaction
            for transaction in client.transactions(since_date=last_reconciliation_date)
            if transaction.cleared in cleared_allowed
        ]
        accounts = client.accounts()
        ids_to_account = {acc.id: acc for acc in accounts}

    if not transactions_to_reconcile:
        print("[bold gree]All accounts are already reconciled!")
        return

    reconcile_groups = display.reconciliation_table(
        ids_to_account, transactions_to_reconcile
    )

    selection = indexes_to_reconcile(max=len(reconcile_groups))

    if selection == "all":
        selected_transactions = transactions_to_reconcile
        selected_ids = {t.account_id for t in selected_transactions}
        selected_accounts = [ids_to_account[acc_id].name for acc_id in selected_ids]
    else:
        selected_transactions = []
        selected_accounts = []
        for index in selection:
            selected_accounts.append(reconcile_groups[index - 1].account_name)
            selected_transactions.extend(reconcile_groups[index - 1].transactions)

    if not selected_transactions:
        print("No accounts to reconcile.\n👋 Bye!")
        return

    print("\nThe following accounts will be reconciled:")
    for acc in selected_accounts:
        print(f"- {acc}")

    if not Confirm.ask("\nShould I go ahead and reconcile them?"):
        print("Alright, cancelling reconciliation.\n👋 Bye!")
        return

    for transaction in selected_transactions:
        transaction.cleared = TransactionClearedStatus.RECONCILED

    with Status("Updating transactions"):
        client.update_transactions(selected_transactions)

    latest_date = max(t.var_date for t in selected_transactions)
    ctx.config.last_reconciliation_date = latest_date - dt.timedelta(
        days=TRANSACTION_GRACE_PERIOD_DAYS
    )
    ctx.config.save()

    print("[bold green]🎉 Reconciliation done!")



import datetime as dt

from ynab.api.accounts_api import AccountsApi
from ynab.api.budgets_api import BudgetsApi
from ynab.api.payees_api import PayeesApi
from ynab.api.transactions_api import TransactionsApi
from ynab.api_client import ApiClient
from ynab.configuration import Configuration
from ynab.models.account import Account
from ynab.models.budget_summary import BudgetSummary
from ynab.models.new_transaction import NewTransaction
from ynab.models.patch_transactions_wrapper import PatchTransactionsWrapper
from ynab.models.payee import Payee
from ynab.models.post_transactions_wrapper import PostTransactionsWrapper
from ynab.models.save_transaction_with_id_or_import_id import (
    SaveTransactionWithIdOrImportId,
)
from ynab.models.transaction_detail import TransactionDetail

from ynab_unlinked.config import Config
from ynab_unlinked.models import TransactionWithYnabData


class Client:
    def __init__(self, config: Config):
        self.config = config
        self.__client = ApiClient(Configuration(access_token=config.api_key))

    def budgets(self, include_accounts: bool = False) -> list[BudgetSummary]:
        api = BudgetsApi(self.__client)
        response = api.get_budgets(include_accounts=include_accounts)
        return response.data.budgets

    def accounts(self) -> list[Account]:
        api = AccountsApi(self.__client)
        response = api.get_accounts(self.config.budget_id)
        return response.data.accounts

    def transactions(
        self, account_id: str | None = None, since_date: dt.date | None = None
    ) -> list[TransactionDetail]:
        api = TransactionsApi(self.__client)

        if account_id:
            response = api.get_transactions_by_account(
                budget_id=self.config.budget_id,
                account_id=account_id,
                since_date=since_date,
            )
        else:
            response = api.get_transactions(
                budget_id=self.config.budget_id,
                since_date=since_date,
            )

        return response.data.transactions

    def payees(self) -> list[Payee]:
        api = PayeesApi(self.__client)
        response = api.get_payees(self.config.budget_id)
        return response.data.payees

    def create_transactions(
        self, account_id: str, transactions: list[TransactionWithYnabData]
    ):
        if not transactions:
            return

        api = TransactionsApi(self.__client)

        transactions_to_create = [
            NewTransaction(
                account_id=account_id,
                date=t.date,
                payee_name=t.payee,
                cleared=t.cleared,
                amount=int(t.amount * 1000),
                approved=False,
                import_id=t.id,
            )
            for t in transactions
        ]

        api.create_transaction(
            self.config.budget_id,
            data=PostTransactionsWrapper(transactions=transactions_to_create),
        )

    def update_transactions(self, transactions: list[TransactionDetail]):
        api = TransactionsApi(self.__client)

        to_update = [
            SaveTransactionWithIdOrImportId(
                id=t.id, account_id=t.account_id, cleared=t.cleared
            )
            for t in transactions
        ]
        api.update_transactions(
            budget_id=self.config.budget_id,
            data=PatchTransactionsWrapper(transactions=to_update),
        )



import datetime as dt
import re
from pathlib import Path

from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.models import Transaction

ANCHOR_LINE = "FECHA|CONCEPTO|LOCALIDAD|IMPORTE"
TRANSACTION_PATTER = re.compile(r"^(\d{2}/\d{2})\|(.+?)\|.+?\|(.*EUR).*")


class SabadellParser:
    def parse(
        self, input_file: Path, context: YnabUnlinkedContext
    ) -> list[Transaction]:
        lines = input_file.read_text(encoding="cp1252").splitlines()
        start = False
        transactions: list[Transaction] = []
        for line in lines:
            if ANCHOR_LINE in line:
                start = True
                continue

            if not start:
                continue

            if groups := TRANSACTION_PATTER.match(line):
                transactions.append(
                    Transaction(
                        date=self.__parse_date(groups[1]),
                        payee=self.__parse_payee(groups[2]),
                        amount=-self.__parse_amount(groups[3]),
                    )
                )
            else:
                start = False

        return transactions

    def __parse_date(self, raw: str) -> dt.date:
        current_year = dt.date.today().year
        return dt.datetime.strptime(f"{raw}/{current_year}", "%d/%m/%Y").date()

    def __parse_payee(self, raw: str) -> str:
        return raw.title()

    def __parse_amount(self, raw: str) -> float:
        return float(raw.replace("EUR", "").replace(",", "."))

    def name(self) -> str:
        return "sabadell"



from pathlib import Path
from typing import Annotated

import typer

from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.process import process_transactions

from .cobee import Cobee, CobeeContext, Language


def command(
    context: typer.Context,
    input_file: Annotated[
        Path,
        typer.Argument(exists=True, file_okay=True, dir_okay=False, readable=True),
    ],
    language: Annotated[
        Language,
        typer.Option(
            "-l",
            "--lang",
            help="Language used when exporting the HTML webisite",
        ),
    ] = Language.ES,
):
    """
    Inputs transactions from a Cobee HTML file.

    Cobee that not support exporting transactions. To use the Cobee entity, form the website,
    find the list of transactions of the month you want to import into YNAB and save the page as HTML.

    You cando this by Right Click > Save as, and select where you want to save the file.
    """

    ctx: YnabUnlinkedContext = context.obj
    ctx.extras = CobeeContext(language=language)

    process_transactions(
        entity=Cobee(),
        input_file=input_file,
        context=ctx,
    )



from enum import StrEnum
from typing import Annotated, assert_never

import typer
from rich import print

from ynab_unlinked.config import Config, ensure_config
from ynab_unlinked.display import prompt_for_api_key, prompt_for_budget


class ValidKeys(StrEnum):
    BUDGET = "budget"
    API_KEY = "api_key"


config = typer.Typer(help="Manage YNAB Unlinked configuration")


@config.command(name="set")
def set_command(
    key: Annotated[
        ValidKeys, typer.Argument(help="The config key to set", show_default=False)
    ],
):
    """Set configuration options"""

    match key:
        case ValidKeys.API_KEY:
            api_key = prompt_for_api_key()
            # If the config exist, just update it
            if ensure_config():
                config = Config.load()
                config.api_key = api_key
            else:
                config = Config(api_key=api_key, budget_id="")

            config.save()
            print("[bold green]🎉 The API key has been updated[/]")
        case ValidKeys.BUDGET:
            budget_id = prompt_for_budget()
            config = Config.load()
            config.budget_id = budget_id
            config.save()
            print("[bold green]🎉 The budget has been updated[/]")
        case never:
            assert_never(never)


@config.command(name="show")
def show():
    if not ensure_config():
        raise typer.Abort(
            "YNAB Unlinked config not found. Run 'yul setup' to configure it."
        )

    config = Config.load()
    print(config.model_dump_json(indent=2))



import datetime as dt
import re
from dataclasses import dataclass
from enum import StrEnum
from pathlib import Path
from typing import assert_never

from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.models import Transaction

DATE_REGEX = re.compile(r"(\d{1,2}) (\w{3}) (\d{4})")

# Needed to parse the date. Cobee is a Spanish system for work benefits but the
# dates are in English locale. Parsing dates with something other than numbers
# does not seem to be supported in Babel.
MONTHS_NUMBERS = {
    "Jan": 1,
    "Feb": 2,
    "Mar": 3,
    "Apr": 4,
    "May": 5,
    "Jun": 6,
    "Jul": 7,
    "Aug": 8,
    "Sep": 9,
    "Oct": 10,
    "Nov": 11,
    "Dec": 12,
}


class Language(StrEnum):
    ES = "es"
    EN = "en"
    PT = "pt"


@dataclass
class CobeeContext:
    language: Language


@dataclass
class Identifiers:
    transactions_line: str
    rejected: str
    cancelled: str


def parse_date(date_str: str) -> dt.date | None:
    if (groups := DATE_REGEX.match(date_str)) is None:
        return None

    day, month, year = groups.groups()

    if month not in MONTHS_NUMBERS:
        raise ValueError(f"The month {month} is not valid.")

    month = MONTHS_NUMBERS[month]

    return dt.date(int(year), int(month), int(day))


def identifers_by_language(language: Language) -> Identifiers:
    match language:
        case Language.ES:
            return Identifiers(
                transactions_line="Transacciones",
                rejected="Rechazada",
                cancelled="Anulada",
            )
        case Language.EN:
            return Identifiers(
                transactions_line="Transactions",
                rejected="Rejected",
                cancelled="Cancelled",
            )
        case Language.PT:
            return Identifiers(
                transactions_line="Transações",
                rejected="Recusada",
                cancelled="Anulada",
            )
        case never:
            assert_never(never)


class Cobee:
    def parse(
        self, input_file: Path, context: YnabUnlinkedContext[CobeeContext]
    ) -> list[Transaction]:
        # Import now the html parser
        import html_text

        text = html_text.extract_text(input_file.read_text())  # type: ignore

        start = False
        previous_line = ""
        transactions: list[Transaction] = []
        date: dt.date | None = None
        payee: str | None = None
        amount: float | None = None
        identifiers = identifers_by_language(context.extras.language)

        for line in text.splitlines():
            line = line.strip()

            if not line:
                continue

            if identifiers.transactions_line in line:
                start = True
                continue

            if not start:
                continue

            if (try_date := parse_date(line)) is not None:
                date = try_date

            if "€" in line:
                amount_str = line.replace("€", "").replace(",", ".")

                try:
                    amount = float(amount_str)
                except ValueError:
                    # If we could not convert this to float it means this is not an amount line.
                    continue

                # If it was the amount line, the previous line is the payee.
                payee = previous_line

                if date is None or payee is None:
                    raise ValueError(
                        f"The input file is not valid. The amount {amount} has been found without a date or payee."
                    )

                transactions.append(Transaction(date=date, payee=payee, amount=amount))
                continue

            if identifiers.cancelled in line or identifiers.rejected in line:
                # These are transactions that didn't went through.
                transactions.pop()
                continue

            previous_line = line

        return transactions

    def name(self) -> str:
        return "cobee"



from __future__ import annotations

import datetime as dt
from pathlib import Path

from pydantic import BaseModel, Field

from ynab_unlinked.models import Transaction, TransactionWithYnabData

CONFIG_PATH = Path.home() / ".config/ynab_unlinked/config.json"
TRANSACTION_GRACE_PERIOD_DAYS = 2


class Checkpoint(BaseModel):
    latest_date_processed: dt.date
    latest_transaction_hash: int


class EntityConfig(BaseModel):
    account_id: str
    checkpoint: Checkpoint | None = None


class Config(BaseModel):
    api_key: str
    budget_id: str
    last_reconciliation_date: dt.date | None = None
    entities: dict[str, EntityConfig] = Field(default_factory=dict)
    payee_rules: dict[str, set[str]] = Field(default_factory=dict)

    def save(self):
        CONFIG_PATH.write_text(self.model_dump_json(indent=4))

    def update_and_save(self, last_transaction: Transaction, entity_name: str):
        checkpoint = Checkpoint(
            latest_date_processed=(
                last_transaction.date - dt.timedelta(days=TRANSACTION_GRACE_PERIOD_DAYS)
            ),
            latest_transaction_hash=hash(last_transaction),
        )

        self.entities[entity_name].checkpoint = checkpoint

        self.save()

    @staticmethod
    def load() -> Config:
        return Config.model_validate_json(CONFIG_PATH.read_text())

    def add_payee_rules(self, transactions: list[TransactionWithYnabData]):
        # For each transaction, add a rule that matches both payees
        for transaction in transactions:
            if transaction.partial_match is None:
                continue

            if transaction.ynab_payee is None:
                continue

            imported_payee = transaction.payee
            ynab_payee = transaction.ynab_payee

            if imported_payee == ynab_payee:
                continue

            self.payee_rules.setdefault(ynab_payee, set()).add(imported_payee)
            self.save()

    def payee_from_fules(self, payee: str) -> str | None:
        return next(
            (
                ynab_payee
                for ynab_payee, valid_names in self.payee_rules.items()
                if payee in valid_names
            ),
            None,
        )


def ensure_config():
    if not CONFIG_PATH.is_file():
        CONFIG_PATH.parent.mkdir(parents=True, exist_ok=True)
        return False
    return True



import typer
from rich import print

from ynab_unlinked import app
from ynab_unlinked.commands import config, load
from ynab_unlinked.config import Config, ensure_config
from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.display import prompt_for_api_key, prompt_for_budget

app.add_typer(load, name="load")
app.add_typer(config, name="config")


@app.command(name="setup")
def setup_command():
    """Setup YNAB Unlinked"""
    print("[bold]Welcome to ynab-unlinked! Lets setup your connection")
    api_key = prompt_for_api_key()
    config = Config(api_key=api_key, budget_id="")
    config.save()

    budget_id = prompt_for_budget()
    config.budget_id = budget_id
    config.save()

    print("[bold green]All done!")


@app.callback(no_args_is_help=True)
def cli(context: typer.Context):
    """
    Create transations in your YNAB account from a bank export of your extract.
    \n

    The first time the command is run you will be asked some questions to setup your YNAB connection. After that,
    transaction processing won't require any input unless there are some actions to take for specific transactions.
    """

    if context.invoked_subcommand == "setup":
        # If we are running setup there is nothing to do here
        return

    if not ensure_config():
        setup_command()
    config = Config.load()
    context.obj = YnabUnlinkedContext(config=config, extras=None)


def main():
    app(prog_name="yul")


if __name__ == "__main__":
    main()



import datetime as dt
from collections.abc import Generator
from pathlib import Path

import typer
from rich import print
from rich.prompt import Confirm, Prompt
from rich.status import Status

from ynab_unlinked import display
from ynab_unlinked.config import (
    TRANSACTION_GRACE_PERIOD_DAYS,
    Checkpoint,
    Config,
    EntityConfig,
)
from ynab_unlinked.context_object import YnabUnlinkedContext
from ynab_unlinked.entities import Entity
from ynab_unlinked.exceptions import ParsingError
from ynab_unlinked.matcher import match_transactions
from ynab_unlinked.models import MatchStatus, Transaction, TransactionWithYnabData
from ynab_unlinked.payee import set_payee_from_ynab
from ynab_unlinked.ynab_api.client import Client

# Request transactions to the YNAB API from the last checkpoint date minus 10 days for buffer
TRANSACTIONS_DAYES_BEFORE_LAST_EXTRACTION = 10


def add_past_to_transactions(
    transactions: list[Transaction], checkpoint: Checkpoint | None
):
    if checkpoint is None:
        return

    for t in transactions:
        if t.date < checkpoint.latest_date_processed + dt.timedelta(
            days=TRANSACTION_GRACE_PERIOD_DAYS
        ):
            t.past = True


def add_counter_to_existing_transactions(transactions: list[Transaction]):
    """
    This method check every transaction that has the same date, payee and amount
    and increments its counter to ensure that they have a unique import ID when
    being added to YNAB.
    """
    counters: dict[str, int] = {}
    for t in transactions:
        if t.id not in counters:
            counters[t.id] = 0
        else:
            counters[t.id] += 1
            t.counter = counters[t.id]


def preprocess_transactions(
    transactions: list[Transaction], checkpoint: Checkpoint | None
):
    add_past_to_transactions(transactions, checkpoint)
    add_counter_to_existing_transactions(transactions)


def filter_transactions(
    transactions: list[Transaction], checkpoint: Checkpoint | None
) -> Generator[Transaction]:
    if checkpoint is None:
        yield from transactions
        return

    yield from (t for t in transactions if t.date >= checkpoint.latest_date_processed)


def get_or_prompt_account_id(config: Config, entity_name: str) -> str:
    if entity_name in config.entities:
        return config.entities[entity_name].account_id

    print(f"Lets select the account for {entity_name.capitalize()}:")
    client = Client(config)

    accounts = [acc for acc in client.accounts() if not acc.closed]

    for idx, acc in enumerate(accounts):
        print(f" - {idx + 1}. {acc.name}")

    acc_num = Prompt.ask(
        "What account are the transactions going to be imported to? (By number)",
        choices=[str(i) for i in range(1, len(accounts) + 1)],
        show_choices=False,
    )
    account = accounts[int(acc_num) - 1]

    print(f"[bold]Account selected: {account.name}")

    config.entities[entity_name] = EntityConfig(account_id=account.id)
    config.save()

    return account.id


def process_transactions(
    entity: Entity,
    input_file: Path,
    context: YnabUnlinkedContext,
) -> None:
    """
    Process the transactions from the input file, match them with YNAB data, and upload them to YNAB.

    If the Entity calling this method does not have a config stored, the user will be prompted to select an account
    this entity should publish transactions to. This account will be used moving forward when using this entity.

    If the user called `yul -a` the user will always be promptped to select
    and account and the selected account won't be saved for this particular entity.
    """

    config = context.config
    show = context.show
    reconcile = context.reconcile

    acount_id = get_or_prompt_account_id(config, entity.name())

    try:
        parsed_input = entity.parse(input_file, context)
    except ParsingError as e:
        print(f"[bold red] Error when parsing {e.input_file}")
        print(f"  Message: {e.message}")
        raise typer.Exit(1) from e

    checkpoint = config.entities[entity.name()].checkpoint

    preprocess_transactions(parsed_input, checkpoint)

    if show:
        display.transaction_table(parsed_input)
        return

    transactions = [
        TransactionWithYnabData(t)
        for t in filter_transactions(
            parsed_input,
            checkpoint,
        )
    ]

    client = Client(config)

    with Status("Reading transactions..."):
        ynab_transactions = client.transactions(
            account_id=acount_id,
            since_date=(
                checkpoint.latest_date_processed
                - dt.timedelta(days=TRANSACTIONS_DAYES_BEFORE_LAST_EXTRACTION)
                if checkpoint
                else None
            ),
        )
    print("[bold green]✔ Transactions read")

    with Status("Augmenting transactions..."):
        match_transactions(transactions, ynab_transactions, reconcile, config)
        set_payee_from_ynab(transactions, client, config)
    print("[bold green]✔ Transactions augmneted with YNAB information")

    display.transactions_to_upload(transactions)

    if not any(t.needs_creation for t in transactions):
        print("[bold blue]🎉 All done! Nothing to do.")
        config.update_and_save(transactions[0], entity.name())
        return

    if partial_matches := [
        t
        for t in transactions
        if t.match_status is MatchStatus.PARTIAL_MATCH and t.needs_creation
    ]:
        display.partial_matches(partial_matches)
        print(
            "\nIf these partial matches are ok, you can accept them and we will keep track of the "
            "payee name for future reference."
        )
        final_matching = (
            MatchStatus.MATCHED
            if Confirm.ask("Do you want to accept these matches?")
            else MatchStatus.UNMATCHED
        )
        for t in partial_matches:
            t.match_status = final_matching
            if final_matching == MatchStatus.UNMATCHED:
                t.reset_matching()

        if final_matching == MatchStatus.MATCHED:
            # If the user agreed to match these, add renaming rules to config
            config.add_payee_rules(partial_matches)

    with Status("Preparing transactions to upload..."):
        new_transactions = [t for t in transactions if t.needs_creation]

    print(f"[bold]Transactions to import:       {len(new_transactions)}")

    if Confirm.ask("Do you want to continue and create the transactions?"):
        with Status("Creating/Updating transactions..."):
            client.create_transactions(acount_id, new_transactions)

        config.update_and_save(transactions[0], entity.name())

    print("[bold blue]🎉 All done!")



from dataclasses import dataclass

from ynab_unlinked.config import Config


@dataclass
class YnabUnlinkedContext[T]:
    config: Config
    extras: T
    show: bool = False
    reconcile: bool = False



from itertools import groupby
from typing import NamedTuple

from rich import box, print
from rich.prompt import Prompt
from rich.rule import Rule
from rich.status import Status
from rich.style import Style
from rich.table import Column, Table
from ynab.models.account import Account
from ynab.models.transaction_detail import TransactionDetail

from ynab_unlinked.config import Config, ensure_config
from ynab_unlinked.models import MatchStatus, Transaction, TransactionWithYnabData
from ynab_unlinked.ynab_api.client import Client

MAX_PAST_TRANSACTIONS_SHOWN = 3


class YnabAmountFormatted(NamedTuple):
    inflow: str
    outflow: str
    amount: str


def format_ynab_amount(ynab_amount: int) -> YnabAmountFormatted:
    value = ynab_amount / 1000
    value_str = f"{value:.2f}€"
    outflow = value_str if ynab_amount < 0 else ""
    inflow = value_str if ynab_amount > 0 else ""

    return YnabAmountFormatted(inflow=inflow, outflow=outflow, amount=value_str)


def prompt_for_api_key() -> str:
    return Prompt.ask("What is the API Key to connect to YNAB?", password=True)


def prompt_for_budget() -> str:
    if ensure_config():
        client = Client(Config.load())
    else:
        api_key = prompt_for_api_key()
        config = Config(api_key=api_key, budget_id="")
        config.save()
        client = Client(config)

    with Status("Getting budgets..."):
        budgets = client.budgets()

    print("Available budgets:")
    for idx, budget in enumerate(budgets):
        print(f" - {idx + 1}. {budget.name}")

    budget_num = Prompt.ask(
        "What budget do you want to use? (By number)",
        choices=[str(i) for i in range(1, len(budgets) + 1)],
        show_choices=False,
    )
    budget = budgets[int(budget_num) - 1]

    print(f"[bold]Selected budget: {budget.name}")
    return budget.id


def transaction_table(transactions: list[Transaction]):
    columns = [
        Column(header="Date", justify="left", max_width=10),
        Column(header="Payee", justify="left", width=50),
        Column(header="Inflow", justify="right", max_width=15),
        Column(header="Outflow", justify="right", max_width=15),
    ]
    table = Table(
        *columns,
        title="Transactions to process",
        caption=f"Only {MAX_PAST_TRANSACTIONS_SHOWN} processed transactions are shown.",
        box=box.SIMPLE,
    )

    past_counter = 0
    for transaction in transactions:
        style = Style(color="gray37" if transaction.past else "default")

        past_counter += int(transaction.past)
        if past_counter == MAX_PAST_TRANSACTIONS_SHOWN:
            # Stop adding transactions that are past after 5 for clarification
            table.add_row("...", "...", "...", "...")
            break

        outflow = transaction.pretty_amount if transaction.amount < 0 else None
        inflow = transaction.pretty_amount if transaction.amount > 0 else None

        table.add_row(
            transaction.date.strftime("%m/%d/%Y"),
            transaction.payee,
            inflow,
            outflow,
            style=style,
        )

    print(table)


def payee_line(transaction: TransactionWithYnabData) -> str:
    if (
        transaction.ynab_payee is not None
        and transaction.payee == transaction.ynab_payee
    ):
        return transaction.ynab_payee

    return f"{transaction.ynab_payee} [gray37] [Original payee: {transaction.payee}][/gray37]"


def updload_help_message(with_partial_matches=False) -> str:
    main_message = (
        "The table below shows the transactaions to be imported to YNAB. The transactions in the input file "
        "have been matched with existing transactions in YNAB.\n"
        " - The [green]green[/] rows are new transactions to be imported.\n"
    )
    if with_partial_matches:
        main_message += (
            " - The [yellow]yellow[/] rows are transaction to be imported that match in date and amount with\n"
            "   transations that exist in YNAB but for which teh payee name could not be matched.\n"
            "   This is usually because the name from the import file is substantially different any payee "
            "present in YNAB.\n"
            "   If you accept these transactions are valid, we will keep track of this naming for future imports."
        )

    main_message += (
        "The cleared status column shows how the transaction will be loaded to YNAB, not the current "
        "status if the transaction was already in YNAB."
    )

    return main_message


def transactions_to_upload(transactions: list[TransactionWithYnabData]):
    columns = [
        Column(header="Match", justify="center", width=5),
        Column(header="Date", justify="left", max_width=10),
        Column(header="Payee", justify="left", width=70),
        Column(header="Inflow", justify="right", max_width=15),
        Column(header="Outflow", justify="right", max_width=15),
        Column(header="Cleared Status", justify="left", width=15),
    ]
    table = Table(
        *columns,
        title="Recent Transactions",
        caption="Transactions to [cyan bold]update[/] and [bold green]create[/].",
        box=box.SIMPLE,
    )

    partial_matches = False
    for transaction in transactions:
        outflow = transaction.pretty_amount if transaction.amount < 0 else None
        inflow = transaction.pretty_amount if transaction.amount > 0 else None

        if transaction.needs_creation:
            if transaction.match_status == MatchStatus.PARTIAL_MATCH:
                style = "yellow"
                partial_matches = True
            else:
                style = "green"
        else:
            style = "default"

        table.add_row(
            transaction.match_emoji,
            transaction.date.strftime("%m/%d/%Y"),
            payee_line(transaction),
            inflow,
            outflow,
            transaction.cleared_status,
            style=style,
        )

    print(Rule("Transactions to be imported"))
    print(updload_help_message(partial_matches))
    print(table)


def partial_matches(transactions: list[TransactionWithYnabData]):
    columns = [
        Column(header="Date", justify="left", max_width=10),
        Column(header="Payee", justify="left", width=50),
        Column(header="Inflow", justify="right", max_width=15),
        Column(header="Outflow", justify="right", max_width=15),
        Column(header="Cleared Status", justify="left", width=15),
    ]
    table = Table(
        *columns,
        title="Partial Matches",
        caption=(
            "Each pair of transactions shows the imported transaction (top) \n"
            "and the partial match in YNAB (bottom)."
        ),
        box=box.SIMPLE,
        row_styles=["", "gray70"],
    )

    for transaction in transactions:
        # If we do not need to import it, skip it
        if not transaction.needs_creation:
            continue

        # Skip if no partial match
        if (
            transaction.match_status != MatchStatus.PARTIAL_MATCH
            or transaction.partial_match is None
        ):
            continue

        # Original transaction row
        orig_outflow = transaction.pretty_amount if transaction.amount < 0 else None
        orig_inflow = transaction.pretty_amount if transaction.amount > 0 else None

        # YNAB transaction row (from partial_match)
        amount = format_ynab_amount(transaction.partial_match.amount)

        # Add the pair of rows
        table.add_row(
            transaction.date.strftime("%m/%d/%Y"),
            transaction.payee,
            orig_inflow,
            orig_outflow,
            "",
        )

        table.add_row(
            transaction.partial_match.var_date.strftime("%m/%d/%Y"),
            transaction.partial_match.payee_name or "",
            amount.inflow,
            amount.inflow,
            transaction.partial_match.cleared.name.capitalize(),
            end_section=True,
        )

    print(table)


class ReconciliationGroup(NamedTuple):
    account_name: str
    transactions: list[TransactionDetail]


def reconciliation_table(
    id_to_account: dict[str, Account], transactions: list[TransactionDetail]
) -> list[ReconciliationGroup]:
    """
    Print a table with the transactions to reconcile per account and return a list
    of transaction groups as displayed in the table.

    The first table is referenced as number one.
    """
    sorted_transactions = sorted(transactions, key=lambda t: t.account_id)

    groups = []
    for counter, (account_id, transaction_group) in enumerate(
        groupby(sorted_transactions, key=lambda t: t.account_id), start=1
    ):
        account = id_to_account[account_id]

        cleared_balance = format_ynab_amount(account.cleared_balance).amount
        uncleared_balance = format_ynab_amount(account.uncleared_balance).amount
        balance = format_ynab_amount(account.balance).amount
        account_name = account.name

        columns = [
            Column(header="Date", justify="left", max_width=10),
            Column(header="Payee", justify="left", width=70),
            Column(header="Inflow", justify="right", max_width=15),
            Column(header="Outflow", justify="right", max_width=15),
            Column(header="Cleared Status", justify="left", width=15),
        ]
        print(
            Rule(
                title=(
                    f"[{counter}] "
                    f"{account_name} - [Balance: [green]{balance}[/] = "
                    f"[green]{cleared_balance}[/] (cleraed) + "
                    f"{uncleared_balance} (uncleared)]"
                ),
                align="left",
                style="bold blue",
            )
        )
        table = Table(*columns)

        group = []
        for transaction in transaction_group:
            group.append(transaction)

            amount = format_ynab_amount(transaction.amount)

            table.add_row(
                transaction.var_date.strftime("%m/%d/%Y"),
                transaction.payee_name,
                amount.inflow,
                amount.outflow,
                TransactionWithYnabData.cleared_str(transaction.cleared),
            )

        print(table)
        print("\n")

        groups.append(
            ReconciliationGroup(account_name=account_name, transactions=group)
        )
    return groups
